/* Kernel module to control the rate
 *
 * Jrme de Vivie   <devivie@info.enserb.u-bordeaux.fr>
 * Herv Eychenne   <eychenne@info.enserb.u-bordeaux.fr>
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include "ip_tables.h"
#include "ipt_RATE.h"

#include <asm/spinlock.h>
#define IPT_RATE_SRC  0x01
#define IPT_RATE_DST  0x02

struct ipt_kern_rate {
	unsigned long rate, prev;
	unsigned long avg, peak;
	unsigned short attack;
	u_int32_t src,dst;
	spinlock_t  lock;
};

#define MS2JIFFIES(ms) (((unsigned int)(ms)) * HZ / 1000)

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

static void
ipt_rate_log(const unsigned int flags, 
	     const unsigned int pos, const struct iphdr *iph)
{
	switch(flags) {
	case IPT_RATE_DST|IPT_RATE_SRC:
		printk("FLOOD: %u.%u.%u.%u->%u.%u.%u.%u, rule %u\n",
		       IP_PARTS(iph->saddr),IP_PARTS(iph->daddr),pos);
		break;
		
	case IPT_RATE_SRC:
		printk("FLOOD: possible scan host from %u.%u.%u.%u, rule %u\n",
		       IP_PARTS(iph->daddr),pos);
		break;
		
	case IPT_RATE_DST:
		printk("FLOOD: possible src spoofed, dst=%u.%u.%u.%u, rule %u\n",
		       IP_PARTS(iph->daddr),pos);
		break;
		
	default:
		printk("FLOOD: random, rule %u\n",pos);
		break;
	}
	return;
}

/* The algorithm used is the Simple Token Bucket Filter (TBF)
 * see net/sched/sch_tbf.c in the linux source tree
 */
static int 
ipt_rate_target(struct sk_buff *skb,
		const char *indev,
		const char *outdev,
		const union ipt_targinfo *info,
		unsigned int pos)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)info->other;
	struct ipt_kern_rate *kr = r->kr;
	struct iphdr *iph= skb->nh.iph;
	unsigned int cpu_flags,flags = 0;
	unsigned long now = jiffies;
	
	spin_lock_irqsave(&kr->lock,cpu_flags);
	kr->rate += now - xchg(&kr->prev, now);
	if (kr->rate > kr->avg) {
		kr->rate = kr->avg;
		kr->attack = 0;
	}

	if (kr->rate >= r->peak) {
		kr->rate -= kr->peak;
		spin_unlock_irqrestore(&kr->lock,cpu_flags);
		return pos + 1;
	}

	if (r->log) {
		switch (kr->attack) {
		case 2:
			goto drop;
		case 1:
			kr->src = iph->saddr;
			kr->dst = iph->daddr;
			kr->attack++;
			break;
		case 0:
		  if (net_ratelimit()) {
			  if(kr->src == iph->saddr)
				  flags |= IPT_RATE_SRC;
			  if(kr->dst == iph->daddr)
				  flags |= IPT_RATE_DST;
			  kr->attack++;
			  spin_unlock_irqrestore(&kr->lock,cpu_flags);
			  ipt_rate_log(flags,pos,iph);
			  return -NF_DROP - 1;
		  }
		}

	} /* if(kr->log) */
 drop:	
       	spin_unlock_irqrestore(&kr->lock,cpu_flags);
	return -NF_DROP - 1;
}

static int 
ipt_rate_checkentry(const union ipt_targinfo *info,
				 unsigned int pos,
				 unsigned int finalpos)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)info->other;
	struct ipt_kern_rate *kr;
	
	if(r->peak >= r->avg)
		return 0;
	
	kr = kmalloc(sizeof(struct ipt_kern_rate), GFP_USER);
	if(!kr)
		return 0;
  
	kr->prev = 0;
	kr->attack = 0;
	kr->rate = MS2JIFFIES(r->avg);
	kr->avg = MS2JIFFIES(r->avg);
	kr->peak = MS2JIFFIES(r->peak);
	kr->lock = SPIN_LOCK_UNLOCKED;
	r->kr = kr;

	return 1;
}

static void
ipt_rate_deleteentry(const union ipt_targinfo *info)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)info->other;
	kfree(r->kr);
}


static struct ipt_target ipt_rate_reg
= { NULL, "RATE", ipt_rate_target, ipt_rate_checkentry, ipt_rate_deleteentry, &__this_module };

int init_module(void)
{
	if (ipt_register_target(&ipt_rate_reg))
		return -EINVAL;
	return 0;
}

void cleanup_module(void)
{
	ipt_unregister_target(&ipt_rate_reg);
}
