/* FTP extention for TCP state tracking. */
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include "../../ip_state.h"
#include "../../ip_state_expect.h"
#include "../ip_state_tcp.h"

EXPORT_NO_SYMBOLS;

#define SERVER_STRING "227 Entering Passive Mode ("
#define CLIENT_STRING "PORT "

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

static struct {
	const char *pattern;
	size_t plen;
	char term;
} search[2] = {
	{ CLIENT_STRING, sizeof(CLIENT_STRING) - 1, '\r' },
	{ SERVER_STRING, sizeof(SERVER_STRING) - 1, ')' }
};

static int try_number(const char *data, size_t dlen, u_int32_t array[6],
		      char term)
{
	size_t i;

	/* Keep data pointing at next char. */
	for (i = 0; dlen; dlen--, data++) {
		if (*data >= '0' && *data <= '9') {
			array[i] = array[i]*10 + *data - '0';
		}
		else if (*data == ',')
			i++;
		else {
			/* Unexpected character; true if it's the
			   terminator and we're finished. */
			return (*data == term && i == 5);
		}
	}

	return 0;
}

/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern(const char *data, size_t dlen,
			const char *pattern, size_t plen,
			char term,
			u_int32_t array[6])
{
	if (dlen == 0)
		return 0;

	if (dlen < plen) {
		if (strnicmp(data, pattern, dlen) == 0)
			return -1;
		else return 0;
	}

	if (strnicmp(data, pattern, plen) != 0)
		return 0;

	if (!try_number(data + plen, dlen - plen, array, term))
		return -1;

	return 1;
}

/* FIXME: This should be in userspace.  Later. */
unsigned int established(const struct tcphdr *tcph, size_t len,
			 int dir,
			 struct ip_state *state,
			 enum ip_statetype *statetype,
			 struct ip_state_expect *expect)
{
	/* dlen not negative guarenteed by ip_state_tcp.c */
	const char *data = (const char *)tcph + tcph->doff * 4;
	size_t dlen = len - tcph->doff * 4;
	u_int32_t array[6] = { 0 };

	switch (find_pattern(data, dlen, search[dir].pattern, 
			     search[dir].plen, search[dir].term, array)) {
	case -1: /* partial */
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		DEBUGP("ip_state_tcp_ftp: partial control packet.\n");
		return NF_DROP;

	case 0: /* no match */
		return NF_ACCEPT;
	}

	/* match */
	expect->tuple = ((struct ip_state_tuple)
			 { { state->tuplehash[!dir].tuple.src.ip, {{ 0 }}, 0 },
			   { htonl((array[0] << 24) | (array[1] << 16)
				   | (array[2] << 8) | array[3]),
			     { { htons(array[4] << 8 | array[5]) } },
			     IPPROTO_TCP }});
	DEBUGP("Expect tuple %p: %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
	       &expect->tuple, IP_PARTS(expect->tuple.src.ip), 
	       ntohs(expect->tuple.src.u.tcp.port),
	       IP_PARTS(expect->tuple.dst.ip), 
	       ntohs(expect->tuple.dst.u.tcp.port));
	/* Make sure hole is 0. */
	expect->tuple.src.u.all = 0;

	modreg_prepend(&expect_list, expect);
	*statetype = IPS_ESTABLISHED_SPECIFIES_RELATED + dir;

	return NF_ACCEPT;
}

struct ip_state_tcp_helper ftp = { NULL, __constant_htons(21), established };

int init_module(void)
{
	if (!ip_state_tcp_helper_register(&ftp))
		return -EINVAL;
	return 0;
}

void cleanup_module(void)
{
	ip_state_tcp_helper_unregister(&ftp);
}
